/*******************************************************************************
* Copyright (C) 2004 Vintela, Inc. All rights reserved.
* Copyright (C) 2005 Novell, Inc. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Vintela, Inc., Novell, Inc., nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Vintela, Inc., Novell, Inc., OR THE 
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
* OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/


/**
 * @author Dan Nuffer
 */

#ifndef BLOCXX_CMD_LINE_PARSER_HPP_INCLUDE_GUARD_
#define BLOCXX_CMD_LINE_PARSER_HPP_INCLUDE_GUARD_
#include "blocxx/BLOCXX_config.h"
#include "blocxx/String.hpp"
#include "blocxx/SortedVectorMap.hpp"
#include "blocxx/Array.hpp"
#include "blocxx/Exception.hpp"

namespace BLOCXX_NAMESPACE
{

BLOCXX_DECLARE_APIEXCEPTION(CmdLineParser, BLOCXX_COMMON_API)

/**
 * Do command line parsing.
 *
 * Thread safety: read/write
 * Copy semantics: Value
 * Exception safety: Strong
 */
class BLOCXX_COMMON_API CmdLineParser
{
public:

	/**
	 * Flags that can be used to specify how a command line option is
	 * processing by the CmdLineParser class.
	 */
	enum EArgumentTypeFlag
	{
		/** The option does not take an argument */
		E_NO_ARG,
		/** the option requires an argument */
		E_REQUIRED_ARG,
		/** The option might have an argument */
		E_OPTIONAL_ARG
	};

	/**
	 * Error codes that may be found in a CmdLineParserException
	 */
	enum EErrorCodes
	{
		/** An unknown option was specified */
		E_INVALID_OPTION,
		/**
		 * An option for which argtype == E_REQUIRED_ARG did not have 
		 * an argument
		 */
		E_MISSING_ARGUMENT,
		/** A non-option argument was specified, but they are not allowed */
		E_INVALID_NON_OPTION_ARG,
		/** The option wasn't specified */
		E_MISSING_OPTION          // the option wasn't specified
	};

	/**
	 * The Option structure is used to hold information about a single command line option.
	 */
	struct Option
	{
		/** Unique option id, used to retrieve option values */
		int id;
		/** Short option char.  Set to '\\0' for none. */
		char shortopt;
		/** Long option string.  Set to 0 for none. */
		const char* longopt;
		/** Specifies constraints for the option's argument */
		EArgumentTypeFlag argtype;
		/** 
		 * If argtype == E_OPTIONAL_ARG and no argument is specified, this 
		 * value will be returned.  Set to 0 for none.
		 */
		const char* defaultValue;
		/** Description used by getUsage(). May be 0. */
		const char* description;
	};

	/**
	 * Flags used to control whether or not non option arguments are
	 * allowed on the command line.
	 */
	enum EAllowNonOptionArgsFlag
	{
		/** Non-option arguments are allowed */
		E_NON_OPTION_ARGS_ALLOWED,
		/** Non-option arguments are NOT allowed */
		E_NON_OPTION_ARGS_INVALID
	};

	/**
	 * Construct a new CmdLineParser object.
	 * @param argc Count of pointers in argv.  Pass value from main().
	 * @param argv Arguments.  Pass value from main(). Value is not saved.
	 * @param options An array of Option objects terminated by a final entry 
	 *		that has a '\\0' shortopt && 0 longopt. Value is not saved.
	 * @param allowNonOptionArgs Indicate whether the presense of non-option arguments is an error.
	 * @throws CmdLineParserException if the given command line is invalid.
	 */
	CmdLineParser(int argc, char const* const* const argv, const Option* options, EAllowNonOptionArgsFlag allowNonOptionArgs);

	/**
	 * Retrieve a value that corresponds to a string option id.
	 * @param id The id of the option to retrieve the value for.
	 * @param defaultValue The return value if the option wasn't set.
	 * @return The value of the option, if given, otherwise defaultValue. If the option was specified more than once, the value
	 *  from the last occurence will be returned.
	 */
	String getOptionValue(int id, const char* defaultValue = "") const;
 	
	/**
	 * Retrieve a value that corresponds to an string option id. Throw and
	 * exception if the option was not specified.
	 * @param id The id of the option to retrieve the value for.
	 * @param exceptionMessage If an exception is thrown this string 
	 *		will be used as the exception message.
	 * @return The value of the option. If the option was specified more than once, the value
	 *  from the last occurence will be returned.
	 * @throws CmdLineParserException with code E_MISSING_OPTION if the option wasn't specified.
	 */
	String mustGetOptionValue(int id, const char* exceptionMessage = "") const;
 	
	/**
	 * Retrieve the values from all occurences of a string option.
	 * @param id The id of the option.
	 * @return The value of the option, if given, otherwise an empty StringArray.
	 */
 	StringArray getOptionValueList(int id) const;
 	
	/**
	 * Retrieve all occurences of a string option.
	 * @param id The id of the option.
	 * @param exceptionMessage If an exception is thrown this string will be
	 *		used as the exception message.
	 * @return The values of the option as a string array.
	 * @throws CmdLineParserException with code E_MISSING_OPTION if the option wasn't specified.
	 */
 	StringArray mustGetOptionValueList(int id, const char* exceptionMessage = "") const;
 	
	/**
	 * Retrieve the value of a boolean option or check for the presence of 
	 * string option.
	 * @param id The id of the option.
	 * @return If id refers to a boolean option, the options value will be returned.
	 *		If id refers to a string option, true is returned if it was specified,
	 *		and false is returned if it was not specified.
	 */
	bool isSet(int id) const;
 	
	/**
	 * @return The number of arguments that aren't options (e.g. filenames).
	 */
	size_t getNonOptionCount () const;
 	
	/**
	 * Retrieve a non-option argument.
	 * @param n The 0-based index of the argument.  Valid values are 0 to count()-1.
	 * @return The value of the non-option argument.
	 */
	String getNonOptionArg(size_t n) const;

	/**
	 * Retrieve all non-option arguments.
	 * @return All non-option arguments as a String array.
	 */
	StringArray getNonOptionArgs() const;
	
	/**
	 * Generate a usage string for the options.  e.g.:
	 * "Options:\n"
	 * "  -1, --one                 first description\n"
	 * "  -2, --two [arg]           second description (default is optional)\n"
	 * "  -3, --three <arg>         third description\n"
	 *
	 * The format [arg] is used to represent E_OPTIONAL_ARG options.
	 * The format \<arg\> is used to represent E_REQUIRED_ARG options.
	 *
	 * @param options An array of Option objects terminated by a final entry 
	 *		that has a '\\0' shortopt && 0 longopt.
	 * @param maxColumns Wrap the descriptions so no line of the usage string 
	 *		exceeds the specified number of columns.
	 */
	static String getUsage(const Option* options, unsigned int maxColumns = 80);


private:

#ifdef BLOCXX_WIN32
#pragma warning (push)
#pragma warning (disable: 4251)
#endif

	// key is Option::id, value is the value(s) specified by the user
	typedef SortedVectorMap<int, StringArray> optionsMap_t;
	optionsMap_t m_parsedOptions;
	StringArray m_nonOptionArgs;

#ifdef BLOCXX_WIN32
#pragma warning (pop)
#endif

};

} // end namespace BLOCXX_NAMESPACE

#endif


