/*******************************************************************************
* Copyright (C) 2004 Vintela, Inc. All rights reserved.
* Copyright (C) 2005 Novell, Inc. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Vintela, Inc., Novell, Inc., nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Vintela, Inc., Novell, Inc., OR THE 
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
* OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/

/**
 * @author Bart Whiteley
 * @author Dan Nuffer
 */

#ifndef BLOCXX_CRYPTOGRAPHIC_RANDOM_NUMBER_HPP_INCLUDE_GUARD_
#define BLOCXX_CRYPTOGRAPHIC_RANDOM_NUMBER_HPP_INCLUDE_GUARD_
#include "blocxx/BLOCXX_config.h"
#include "blocxx/Exception.hpp"

namespace BLOCXX_NAMESPACE
{

BLOCXX_DECLARE_APIEXCEPTION(CryptographicRandomNumber, BLOCXX_COMMON_API);

} // end namespace BLOCXX_NAMESPACE

#if defined(BLOCXX_HAVE_OPENSSL)
#include "blocxx/Types.hpp"

#include <stdlib.h> // for RAND_MAX

namespace BLOCXX_NAMESPACE
{

/**
 * The CryptographicRandomNumber class generates "secure" random numbers that
 * should be suitable for cryptograhic processing.
 */
class BLOCXX_COMMON_API CryptographicRandomNumber
{
public:

	/**
	 * Create a CryptographicRandomNumber object that will only produce
	 * random number within a given range.
	 * @param lowVal The lower end of the range for random numbers.
	 * @param highVal The upper end of the range for random numbers.
	 */
	CryptographicRandomNumber(Int32 lowVal = 0, Int32 highVal = RAND_MAX);

	/**
	 * @return The next random number from the CryptographicRandomNumber
	 * 		object.
	 * @throws CryptographicRandomNumberException on failure. The caller 
	 * should treat this as a serious error.
	 */
	Int32 getNextNumber();
	
public:

	/**
	 * This function can be called to control when the openssl cryptographic 
	 * prng will be initialized. It may be a lengthy process on some OSs 
	 * which don't have /dev/urandom. If it hasn't been previously called, it 
	 * will be called the first time a CryptographicRandomNumber instance is 
	 * instantiated. It's okay to call this function multiple times from 
	 * multiple threads.
	 */
	static void initRandomness();

	/**
	 * Save the random state of this CryptographicRandomNumber
	 */
	static void saveRandomState();

private:
	Int32 m_lowVal;
	Int32 m_highVal;
	Int32 m_range;
	int m_numBits;
};

} // end namespace BLOCXX_NAMESPACE

#else

// no openssl, so just typedef the class to RandomNumber
#include "blocxx/RandomNumber.hpp"

namespace BLOCXX_NAMESPACE
{

typedef RandomNumber CryptographicRandomNumber;

} // end namespace BLOCXX_NAMESPACE

#endif // #if defined(BLOCXX_HAVE_OPENSSL)

#endif

