/*******************************************************************************
* Copyright (C) 2005 Novell, Inc. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Novell, Inc., nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Novell, Inc., OR THE 
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
* OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/

/*
    Acknowledgements: this header is heavily based on ScopeGuard.h by Andrei 
	Alexandrescu and Petru Marginean publishedin December 2000 issue of C/C++ 
	Users Journal:
	http://www.cuj.com/documents/s=8000/cujcexp1812alexandr/alexandr.htm

	The only modification from the original form are the following:
	- Inclusion in the BLOCXX_NAMESPACE
	- function/method/member variable renaming to comply with the blocxx
	  coding guidelines.
*/

#ifndef BLOCXX_SCOPEGUARD_HPP_
#define BLOCXX_SCOPEGUARD_HPP_

namespace BLOCXX_NAMESPACE
{

//////////////////////////////////////////////////////////////////////////////
template <class T>
class RefHolder
{
	T& ref_;
public:
	RefHolder(T& ref) : ref_(ref) {}
	operator T& () const 
	{
		return ref_;
	}
private:
    // Disable assignment - not implemented
    RefHolder& operator=(const RefHolder&);
};

template <class T>
inline RefHolder<T> 
byRef(T& t)
{
	return RefHolder<T>(t);
}

//////////////////////////////////////////////////////////////////////////////
class ScopeGuardImplBase
{
	ScopeGuardImplBase& operator =(const ScopeGuardImplBase&);
protected:
	~ScopeGuardImplBase()
	{
	}
	ScopeGuardImplBase(const ScopeGuardImplBase& other) throw() 
		: m_dismissed(other.m_dismissed)
	{
		other.dismiss();
	}
	template <typename J>
	static void safeExecute(J& j) throw() 
	{
		if (!j.m_dismissed)
			try
			{
				j.execute();
			}
			catch(...)
			{
			}
	}
	
	mutable bool m_dismissed;
public:
	ScopeGuardImplBase() throw() : m_dismissed(false) 
	{
	}
	void dismiss() const throw() 
	{
		m_dismissed = true;
	}
};

typedef const ScopeGuardImplBase& ScopeGuard;

//////////////////////////////////////////////////////////////////////////////
template <typename F>
class ScopeGuardImpl0 : public ScopeGuardImplBase
{
public:
	static ScopeGuardImpl0<F> makeGuard(F fun)
	{
		return ScopeGuardImpl0<F>(fun);
	}
	~ScopeGuardImpl0() throw() 
	{
		safeExecute(*this);
	}
	void execute() 
	{
		fun_();
	}
protected:
	ScopeGuardImpl0(F fun) : fun_(fun) 
	{
	}
	F fun_;
};

template <typename F> 
inline ScopeGuardImpl0<F> 
makeGuard(F fun)
{
	return ScopeGuardImpl0<F>::makeGuard(fun);
}

//////////////////////////////////////////////////////////////////////////////
template <typename F, typename P1>
class ScopeGuardImpl1 : public ScopeGuardImplBase
{
public:
	static ScopeGuardImpl1<F, P1> makeGuard(F fun, P1 p1)
	{
		return ScopeGuardImpl1<F, P1>(fun, p1);
	}
	~ScopeGuardImpl1() throw() 
	{
		safeExecute(*this);
	}
	void execute()
	{
		fun_(m_p1);
	}
protected:
	ScopeGuardImpl1(F fun, P1 p1) : fun_(fun), m_p1(p1) 
	{
	}
	F fun_;
	const P1 m_p1;
};

template <typename F, typename P1> 
inline ScopeGuardImpl1<F, P1> 
makeGuard(F fun, P1 p1)
{
	return ScopeGuardImpl1<F, P1>::makeGuard(fun, p1);
}

//////////////////////////////////////////////////////////////////////////////
template <typename F, typename P1, typename P2>
class ScopeGuardImpl2: public ScopeGuardImplBase
{
public:
	static ScopeGuardImpl2<F, P1, P2> makeGuard(F fun, P1 p1, P2 p2)
	{
		return ScopeGuardImpl2<F, P1, P2>(fun, p1, p2);
	}
	~ScopeGuardImpl2() throw() 
	{
		safeExecute(*this);
	}
	void execute()
	{
		fun_(m_p1, m_p2);
	}
protected:
	ScopeGuardImpl2(F fun, P1 p1, P2 p2) : fun_(fun), m_p1(p1), m_p2(p2) 
	{
	}
	F fun_;
	const P1 m_p1;
	const P2 m_p2;
};

template <typename F, typename P1, typename P2>
inline ScopeGuardImpl2<F, P1, P2> 
makeGuard(F fun, P1 p1, P2 p2)
{
	return ScopeGuardImpl2<F, P1, P2>::makeGuard(fun, p1, p2);
}

//////////////////////////////////////////////////////////////////////////////
template <typename F, typename P1, typename P2, typename P3>
class ScopeGuardImpl3 : public ScopeGuardImplBase
{
public:
	static ScopeGuardImpl3<F, P1, P2, P3> makeGuard(F fun, P1 p1, P2 p2, P3 p3)
	{
		return ScopeGuardImpl3<F, P1, P2, P3>(fun, p1, p2, p3);
	}
	~ScopeGuardImpl3() throw() 
	{
		safeExecute(*this);
	}
	void execute()
	{
		fun_(m_p1, m_p2, m_p3);
	}
protected:
	ScopeGuardImpl3(F fun, P1 p1, P2 p2, P3 p3) : fun_(fun), m_p1(p1), m_p2(p2), m_p3(p3) 
	{
	}
	F fun_;
	const P1 m_p1;
	const P2 m_p2;
	const P3 m_p3;
};

template <typename F, typename P1, typename P2, typename P3>
inline ScopeGuardImpl3<F, P1, P2, P3> 
makeGuard(F fun, P1 p1, P2 p2, P3 p3)
{
	return ScopeGuardImpl3<F, P1, P2, P3>::makeGuard(fun, p1, p2, p3);
}

//////////////////////////////////////////////////////////////////////////////
template <class Obj, typename MemFun>
class ObjScopeGuardImpl0 : public ScopeGuardImplBase
{
public:
	static ObjScopeGuardImpl0<Obj, MemFun> makeObjGuard(Obj& obj, MemFun memFun)
	{
		return ObjScopeGuardImpl0<Obj, MemFun>(obj, memFun);
	}
	~ObjScopeGuardImpl0() throw() 
	{
		safeExecute(*this);
	}
	void execute() 
	{
		(m_obj.*m_memFun)();
	}
protected:
	ObjScopeGuardImpl0(Obj& obj, MemFun memFun) 
		: m_obj(obj), m_memFun(memFun) {}
	Obj& m_obj;
	MemFun m_memFun;
};

template <class Obj, typename MemFun>
inline ObjScopeGuardImpl0<Obj, MemFun> 
makeObjGuard(Obj& obj, MemFun memFun)
{
	return ObjScopeGuardImpl0<Obj, MemFun>::makeObjGuard(obj, memFun);
}

//////////////////////////////////////////////////////////////////////////////
template <class Obj, typename MemFun, typename P1>
class ObjScopeGuardImpl1 : public ScopeGuardImplBase
{
public:
	static ObjScopeGuardImpl1<Obj, MemFun, P1> makeObjGuard(Obj& obj, MemFun memFun, P1 p1)
	{
		return ObjScopeGuardImpl1<Obj, MemFun, P1>(obj, memFun, p1);
	}
	~ObjScopeGuardImpl1() throw() 
	{
		safeExecute(*this);
	}
	void execute() 
	{
		(m_obj.*m_memFun)(m_p1);
	}
protected:
	ObjScopeGuardImpl1(Obj& obj, MemFun memFun, P1 p1) 
		: m_obj(obj), m_memFun(memFun), m_p1(p1) {}
	Obj& m_obj;
	MemFun m_memFun;
	const P1 m_p1;
};

template <class Obj, typename MemFun, typename P1>
inline ObjScopeGuardImpl1<Obj, MemFun, P1> 
makeObjGuard(Obj& obj, MemFun memFun, P1 p1)
{
	return ObjScopeGuardImpl1<Obj, MemFun, P1>::makeObjGuard(obj, memFun, p1);
}

//////////////////////////////////////////////////////////////////////////////
template <class Obj, typename MemFun, typename P1, typename P2>
class ObjScopeGuardImpl2 : public ScopeGuardImplBase
{
public:
	static ObjScopeGuardImpl2<Obj, MemFun, P1, P2> makeObjGuard(Obj& obj, MemFun memFun, P1 p1, P2 p2)
	{
		return ObjScopeGuardImpl2<Obj, MemFun, P1, P2>(obj, memFun, p1, p2);
	}
	~ObjScopeGuardImpl2() throw() 
	{
		safeExecute(*this);
	}
	void execute() 
	{
		(m_obj.*m_memFun)(m_p1, m_p2);
	}
protected:
	ObjScopeGuardImpl2(Obj& obj, MemFun memFun, P1 p1, P2 p2) 
		: m_obj(obj), m_memFun(memFun), m_p1(p1), m_p2(p2) {}
	Obj& m_obj;
	MemFun m_memFun;
	const P1 m_p1;
	const P2 m_p2;
};

template <class Obj, typename MemFun, typename P1, typename P2>
inline ObjScopeGuardImpl2<Obj, MemFun, P1, P2> 
makeObjGuard(Obj& obj, MemFun memFun, P1 p1, P2 p2)
{
	return ObjScopeGuardImpl2<Obj, MemFun, P1, P2>::makeObjGuard(obj, memFun, p1, p2);
}

#define CONCATENATE_DIRECT(s1, s2) s1##s2
#define CONCATENATE(s1, s2) CONCATENATE_DIRECT(s1, s2)
#define ANONYMOUS_VARIABLE(str) CONCATENATE(str, __LINE__)

#define ON_BLOCK_EXIT ScopeGuard ANONYMOUS_VARIABLE(scopeGuard) = makeGuard
#define ON_BLOCK_EXIT_OBJ ScopeGuard ANONYMOUS_VARIABLE(scopeGuard) = makeObjGuard

}	// End of BLOCXX_NAMESPACE

#endif //SCOPEGUARD_H_
